# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""SDC Custom S3 Operators"""
import logging
import os
import warnings
from urllib.parse import urlparse

import boto3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


def parse_s3_url(url):
    """Returns an (s3 bucket, key name/prefix) tuple from a url with an s3 scheme.
    Args:
        url (str):
    Returns:
        tuple: A tuple containing:
            - str: S3 bucket name
            - str: S3 key
    """
    parsed_url = urlparse(url)
    if parsed_url.scheme != "s3":
        raise ValueError("Expecting 's3' scheme, got: {} in {}.".format(parsed_url.scheme, url))
    return parsed_url.netloc, parsed_url.path.lstrip("/")


class CopyBetweenS3Locations(BaseOperator):
    template_fields = ('dst_path', 'src_path', 'keep_filename', 'xcom_task_id', 'xcom_task_id_key',)

    @apply_defaults
    def __init__(self, xcom_task_id, xcom_task_id_key, dst_path,
                 src_path=None, keep_filename=False, *args, **kwargs):
        super(CopyBetweenS3Locations, self).__init__(*args, **kwargs)
        self.dst_path = dst_path
        self.src_path = src_path
        self.keep_filename = keep_filename

        self.xcom_task_id = xcom_task_id
        self.xcom_task_id_key = xcom_task_id_key

        if kwargs.get('xcom_task_id', None) is not None:
            warnings.warn("xcom_task_id is not implemented for this operator.")
        if kwargs.get('xcom_task_id_key', None) is not None:
            warnings.warn("xcom_task_id_key is not implemented for this operator.")

    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(CopyBetweenS3Locations, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            Copy s3 object between two buckets
        """
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SM_RESOURCE = BOTO_SESSION.resource('s3')

        try:
            if self.src_path is None:
                self.src_path = context['task_instance'] \
                    .xcom_pull(self.xcom_task_id, key=self.xcom_task_id_key)
                assert self.src_path is not None, \
                    "No source path found. Either hardcode in config or pass from previous job"

            src_bucket, src_filepath = parse_s3_url(self.src_path)
            dst_bucket, dst_filepath = parse_s3_url(self.dst_path)

            if self.keep_filename:
                # assumes destination path is a folder path
                # keeps src filename
                src_filename = src_filepath.split('/')[-1]
                logger.info("Keep source filename = {}".format(src_filename))
                # assumes dst_path given without filename
                dst_filepath = os.path.join(dst_filepath, src_filename)
                # assumes dst_path is given without filename
                self.dst_path = os.path.join(self.dst_path, src_filename)

            copy_source = {
                'Bucket': src_bucket,
                'Key': src_filepath
            }
            bucket = SM_RESOURCE.Bucket(dst_bucket)
            bucket.copy(copy_source, dst_filepath)

            # push response up
            context['task_instance'].xcom_push(
                key='destination_location',
                value=self.dst_path
            )

        except Exception as exception:
            raise exception



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""
    TERRAFORM TASK CODE
        - these are used in pythonOperator steps as a way \
            to flexibly call boto3 and perform any special requirements
"""
import json
import logging

import boto3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class GetRemoteTfstate(BaseOperator):
    """
       A process of collecting the relevant tfstate resources based on the terraform
       tags provided. If tags are not provided, the entire tfstate resources are returned.
    """
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(GetRemoteTfstate, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """

        """
        s3 = boto3.client('s3')

        s3_bucket_name = self.config.get('s3_bucket_name', None)
        s3_bucket_key = self.config.get('s3_bucket_key', None)
        terraform_tags = self.config.get('terraform_tags', None)

        try:
            obj = s3.get_object(
                Bucket=s3_bucket_name, Key=s3_bucket_key
            )

            tfstate_raw_data = json.loads(obj['Body'].read())

            tfstate_resources = tfstate_raw_data.get('resources')
            tfstate_data = []
            for tfstate_instance in tfstate_resources:
                tags = tfstate_instance.get('instances')[0].get('attributes').get('tags', {})
                name = tfstate_instance \
                    .get('instances')[0] \
                    .get('attributes') \
                    .get('name',
                         # here if name does not exist, then the name is found in the id
                         tfstate_instance \
                         .get('instances')[0] \
                         .get('attributes') \
                         .get('id', None)
                         )
                type = tfstate_instance.get('type', None)

                if terraform_tags is not None:
                    if tags.get('client') == terraform_tags.get('client') \
                            and tags.get('service') == terraform_tags.get('service') \
                            and tags.get('environment') == terraform_tags.get('environment'):

                        if terraform_tags.get('use_case') in name:
                            tfstate_data.append({
                                'resource_name': name,
                                'resource_tags': tags,
                                'resource_type': type
                            })

            context['task_instance'].xcom_push(
                key='remote_tfstate',
                value=tfstate_data
            )

        except Exception as err:
            raise err



"""TestExtendedDagFactory"""
from pathlib import Path

import pytest
import yaml

from sdc_orchestration_helpers.airflow.dag_factory import ExtendedDagFactory
from sdc_orchestration_helpers.yaml_helpers.loaders import get_yaml_config


class TestExtendedDagFactory:
    """Test Extended DagFactory class"""

    @staticmethod
    def test_simple_config_load():
        """Test the extended dagfactory loads config as expected"""
        # pylint: disable=protected-access

        simple_config = Path('tests/fixtures/simple_dummy_dag.yml').absolute()

        try:
            dag_factory = ExtendedDagFactory(config_filepath=simple_config, loader=yaml.SafeLoader)
            _ = dag_factory._load_config(config_filepath=simple_config)
        except Exception as exception:
            raise pytest.fail("DID RAISE {0}".format(exception))

    @staticmethod
    def test_simple_config_correct_output():
        """Test the extended dagfactory loads config with expected output"""
        # pylint: disable=protected-access
        simple_config = Path('tests/fixtures/simple_dummy_dag.yml').absolute()
        validation_yaml = get_yaml_config(filepath=simple_config)
        dag_factory = ExtendedDagFactory(config_filepath=simple_config, loader=yaml.SafeLoader)
        actual = dag_factory._load_config(config_filepath=simple_config)

        return actual == validation_yaml, (
            "failed. Loaded config did not match the validation config."
        )



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""
    ATHENA TASK CODE
        - these are used in pythonOperator steps as a way \
            to flexibly call boto3 and perform any special requirements
"""
import logging

import boto3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class StartQueryExecution(BaseOperator):
    """Runs athena query against feature store."""
    template_fields = ('config',)

    @apply_defaults
    def __init__(self, config, *args, **kwargs):
        super(StartQueryExecution, self).__init__(*args, **kwargs)
        self.config = config

    def execute(self, context):
        """
            https://sagemaker.readthedocs.io/en/stable/api/\
                prep_data/feature_store.html#sagemaker.feature_store.feature_group.AthenaQuery
        """
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        ATHENA_CLIENT = BOTO_SESSION.client('athena')

        try:
            response = ATHENA_CLIENT.start_query_execution(
                QueryString=self.config['query'],
                QueryExecutionContext=self.config.get('query_execution_context', {}),
                ResultConfiguration=self.config['result_configuration'],
                WorkGroup=self.config['workgroup']
            )

            context['task_instance'].xcom_push(
                key='athena_query_execution_id',
                value=response['QueryExecutionId']
            )

        except Exception as err:
            raise err


class GetQueryExecution(BaseOperator):
    """Runs athena query against feature store."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(GetQueryExecution, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            https://sagemaker.readthedocs.io/en/stable/api/\
                prep_data/feature_store.html#sagemaker.feature_store.feature_group.AthenaQuery
        """
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        ATHENA_CLIENT = BOTO_SESSION.client('athena')

        try:

            query_execution_id = self.config.get('query_execution_id', None)
            task_ids = self.xcom_tasks['query_execution_id']['task_id']

            if query_execution_id is None:
                query_execution_id = context['task_instance'].xcom_pull(
                    task_ids=task_ids,
                    key=self.xcom_tasks['query_execution_id']['key']
                )

            logging.info("With query_id = {}\n".format(query_execution_id))

            # loop while results completing
            # on complete, do training next
            response = ATHENA_CLIENT.get_query_execution(
                QueryExecutionId=query_execution_id
            )
            query_execution_response = response.get('QueryExecution', {})
            job_state = query_execution_response.get('Status', {}).get('State', None)
            reason = query_execution_response.get('Status', {}).get('StateChangeReason', None)
            output_location = query_execution_response.get(
                'ResultConfiguration', {}).get('OutputLocation', None)
            # push response up
            context['task_instance'].xcom_push(
                key='output_location',
                value=output_location
            )

            if job_state == 'SUCCEEDED':
                return True
            elif job_state in ['QUEUED', 'RUNNING']:
                return False
            else:
                # retry on failure
                raise Exception(
                    "query status == {job_state} with reason = {reason}".format(
                        job_state=job_state, reason=reason
                    )
                )

        except Exception as err:
            raise err



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""
    GLUE TASK CODE
        - these are used in pythonOperator steps as a way \
            to flexibly call boto3 and perform any special requirements
"""
import logging
import time
import warnings

import boto3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class RunGlueJob(BaseOperator):
    """Runs Glue Job, and can consume variables for the job."""
    template_fields = ('config', 'xcom_tasks')

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(RunGlueJob, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            :config.job_name: str. the name of the glue job to run.
            :config.job_arguments: dict. the key value arguments send to the job.
            :xcom_push: str. glue_job_run_id.
        """
        CLIENT = boto3.client('glue')

        job_name = self.config.get('job_name', None)
        job_arguments = self.config.get('job_arguments', None)

        if job_name is None:
            raise ValueError('Job name is required for task.')
        if job_arguments is None:
            warnings.warn('Job arguments are empty.')

        try:
            logging.info('Running glue job {name} with envs {envs}.'.format(
                name=job_name,
                envs=job_arguments)
            )

            response = CLIENT.start_job_run(
                JobName=job_name,
                Arguments=job_arguments
            )

            # push job_run_id to xcom
            context['task_instance'].xcom_push(
                key='job_run_id',
                value=response['JobRunId']
            )

            # push job_name to xcom for other glue tasks to reference
            context['task_instance'].xcom_push(
                key='job_name',
                value=job_name
            )

        except Exception as err:
            raise err


class MonitorGlueJob(BaseOperator):
    """Monitors Glue Job until SUCCESS state."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks, *args, **kwargs):
        super(MonitorGlueJob, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            :config.job_name: str. optional the name of the glue job to monitor.
                             if not monitored, name is pulled from xcom_tasks.
            :config.job_run_id: str. optional run id of the job to monitor
                               if not monitored, id is pulled from xcom_tasks.
            xcom_tasks:
                job_name:
                    task_id: str.
                    key: job_name
                job_run_id:
                    task_id: str
                    key: job_run_id
        """
        CLIENT = boto3.client('glue')

        try:
            job_name = self.config.get('job_name', None)
            job_run_id = self.config.get('job_run_id', None)

            task_ids = self.xcom_tasks['job_name']['task_id']

            if job_name is None:
                # if optional config job_name is None, pull from xcom
                job_name = context['task_instance'].xcom_pull(
                    task_ids=task_ids,
                    key=self.xcom_tasks['job_name']['key']
                )

            if job_run_id is None:
                # if optional config job_run_id is None, pull from xcom
                job_run_id = context['task_instance'].xcom_pull(
                    task_ids=task_ids,
                    key=self.xcom_tasks['job_run_id']['key']
                )

            logging.info('Got glue job name and run id from xcom: {name}:{id}'.format(
                name=job_name, id=job_run_id))

            while True:
                response = CLIENT.get_job_run(
                    JobName=job_name,
                    RunId=job_run_id,
                )

                state = response['JobRun']['JobRunState']
                if state == 'SUCCEEDED':
                    break
                if state == 'STOPPED':
                    break
                if state == 'FAILED':
                    raise RuntimeError('Glue Job failed.', response)
                if state == 'TIMEOUT':
                    raise RuntimeError('Glue Job timed out.')

                logging.info('Monitoring glue job: {name}.\n State: {state}'.format(
                    name=job_name,
                    state=state)
                )

                time.sleep(10)

        except Exception as err:
            raise err


class TagGlueJob(BaseOperator):
    """Adds a tag to the resource specified."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(TagGlueJob, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """

        """
        CLIENT = boto3.client('glue')

        job_name = self.config.get('job_name', None)
        resource_tags = self.config.get('resource_tags', None)

        if job_name is None:
            # if optional config job_name is None, pull from xcom
            job_name = context['task_instance'].xcom_pull(
                self.xcom_tasks['job_name']['task_id'],
                key=self.xcom_tasks['job_name']['key']
            )

        if resource_tags is None:
            raise ValueError('Resource tags are required for task.')

        try:
            account_id = boto3.client('sts').get_caller_identity().get('Account')
            glue_resource_arn = 'arn:aws:glue:eu-west-1:{account_id}:job/{name}'.format(
                account_id=account_id, name=job_name
            )

            CLIENT.tag_resource(
                ResourceArn=glue_resource_arn,
                TagsToAdd=resource_tags
            )

            # push resource_tags to xcom for other glue tasks to reference
            context['task_instance'].xcom_push(
                key='resource_tags',
                value=resource_tags
            )

        except Exception as err:
            raise err


class UnTagGlueJob(BaseOperator):
    """Removes a tag to the resource specified."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(UnTagGlueJob, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            :glue_resource_arn: str. the Arn of the glue resource.
            :resource_tags: dict. the key value pairs of the
                                 relevant tags to remove.
        """
        CLIENT = boto3.client('glue')

        job_name = self.config.get('job_name', None)
        resource_tags = self.config.get('resource_tags', None)

        task_ids = self.xcom_tasks['job_name']['task_id']

        if job_name is None:
            # if optional config job_name is None, pull from xcom
            job_name = context['task_instance'].xcom_pull(
                task_ids=task_ids,
                key=self.xcom_tasks['job_name']['key']
            )

        if resource_tags is None:
            # if optional config resource_tags None, pull from xcom
            resource_tags = context['task_instance'].xcom_pull(
                task_ids=task_ids,
                key=self.xcom_tasks['resource_tags']['key']
            )

        try:
            account_id = boto3.client('sts').get_caller_identity().get('Account')
            glue_resource_arn = 'arn:aws:glue:eu-west-1:{account_id}:job/{name}'.format(
                account_id=account_id, name=job_name
            )

            tags_to_remove = list(key for key in resource_tags.keys())

            CLIENT.untag_resource(
                ResourceArn=glue_resource_arn,
                TagsToRemove=tags_to_remove
            )

        except Exception as err:
            raise err


class RunGlueCrawler(BaseOperator):
    """Runs Glue Crawler."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config=None, xcom_tasks=None, *args, **kwargs):
        super(RunGlueCrawler, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            :glue_crawler_name: str. the name of the crawler to be triggered.
        """
        CLIENT = boto3.client('glue')

        crawler_name = self.config.get('crawler_name', None)
        if crawler_name is None:
            raise ValueError('Crawler name needed for task.')

        try:
            logging.info(
                'Running glue crawler {name}'.format(
                    name=crawler_name
                )
            )

            CLIENT.start_crawler(
                Name=crawler_name
            )

            # push crawler_name to xcom
            context['task_instance'].xcom_push(
                key='crawler_name',
                value=crawler_name
            )

        except Exception as err:
            raise err


class MonitorGlueCrawler(BaseOperator):
    """Monitors Glue Crawler until complete."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(MonitorGlueCrawler, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            :glue_crawler_name: str. the name of crawler to be monitored.
        """
        CLIENT = boto3.client('glue')

        crawler_name = self.config.get('crawler_name', None)
        task_ids = self.xcom_tasks['job_name']['task_id']

        if crawler_name is None:
            crawler_name = context['task_instance'].xcom_pull(
                task_ids=task_ids,
                key=self.xcom_tasks['crawler_name']['key']
            )

        try:
            wait = 0
            while True:
                state = CLIENT.get_crawler(
                    Name=crawler_name
                )['Crawler']['State']

                if state == 'READY':
                    break
                if state == 180:
                    raise RuntimeError('Pre crawler: {name} failed. Timeout.'.format(
                        name=crawler_name)
                    )

                logging.info('Monitoring glue crawler: {name}.\n State: {state}'.format(
                    name=crawler_name, state=state)
                )

                wait += 1
                time.sleep(10)
        except Exception as err:
            raise err
